/*
 * usb_8o.c
 *
 * Created: 2016-01-08 10:26:25
 *  Author: dsosnowski
 */ 

#define F_CPU		8000000UL 


//includes
#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <util/delay.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
//#include <avr/pgmspace.h>

//outputs for relays
#define OUTPORT PORTB 
#define OUT1 	7
#define OUT2 	6
#define OUT3 	5
#define OUT4 	4
#define OUT5 	3
#define OUT6 	2
#define OUT7 	1
#define OUT8 	0

//extrnal signals
#define SIGPORT PORTD
#define RTS		6
#define CTS 	5
#define LED 	4
//#define USB_A 	3
//#define USB_B 	2
#define EXT_RST 	2
#define TXD 	1
#define RXD 	0

//constants
//#define F_CPU			8000000
#define NEW_COMMAND		0x40
#define INIT_OK			0x01
#define HELLO_SIG		6
#define CLEAR			0
#define STOP			0
#define DISABLE			0
#define SET				1
#define ENABLE			1
//on/off echo
#define ECHO			DISABLE

//in ms
#define NOICE_DELAY		5
#define CHANGE_DELAY	200
#define AUTO_CHECK_DELAY 1000
#define MAIN_DELAY		50


//variables
static volatile char command;
static volatile char parametr;
static volatile char parametr2;
static volatile char pulse_index;
static volatile char check_loop;
static volatile char cyclic_check;

//subs
void hard_init(void);
void led_flasher(char pulses);
void led_driver(char state);
void con_active(void);
void rx_control(char state);
void tx_char(char send);
void send_port_state(void);
void send_pin_state(char pin);
void send_port_config(void);
void invalid(void);
char get_digit(char character);
char get_char(char value);


char ext_rst = 0;
char master_reset = 0;

int main(void)
{
	char temp;
	char i;

		
	//------------------------------
	//hard init ok	
	hard_init();
	led_flasher(7);
	send_port_state();

	//------------------------------
	//main
	while(1)
	{
		if (command)
		{
			GIMSK = DISABLE;	//disable pin change
			//con_active();	
			switch (command)
			{
				
			//D out to port
				case 'D':
					temp = get_digit(parametr);
					i = get_digit(parametr2);
					
					if ((temp > 'F') || ( i > 'F'))
					{
						invalid();
					}
					else
					{
						//send_port_state();
						temp <<= 4;
						temp |= i;
						OUTPORT = temp;
					}

				break;

			//R read actual state
				case 'R':
					send_port_state();
				break;

			//P config port directions
				case 'P':
					temp = get_digit(parametr);
					i = get_digit(parametr2);
					
					if ((temp > 'F') || ( i > 'F'))
					{
						invalid();
					}
					else
					{
						//send_port_state();
						temp <<= 4;
						temp |= i;
						//find when before not out and now is out
						for (i = 0; i < 8; i++)
						{
							//if new is out
							if (temp & (1<<i))
							{
								//if not out before	then clear output
								if (!(DDRB & (1<<i))) OUTPORT &= ~(1<<i);
							}
						}
						DDRB = temp;
					}
				break;
				
			//G get actual configutation
				case 'G':
					send_port_config();
				break;

			//L get pin level (state)
				case 'L':
					temp = get_digit(parametr2);
					if ((temp >= 1) && ( temp <= 8))
					{
						temp--;
						send_pin_state(temp);
					}
					else invalid();
				break;

			//S set pin
				case 'S':
					temp = get_digit(parametr2);
					if ((temp >= 1) && ( temp <= 8))
					{
						temp--;
						PORTB |= (1<<temp);
						//send_port_state();
					}
					else invalid();
				break;
			
			//C clear pin
				case 'C':
					temp = get_digit(parametr2);
					if ((temp >= 1) && ( temp <= 8))
					{
						temp--;
						PORTB &= ~(1<<temp);
						//send_port_state();
					}
					else invalid();
				break;
		
				default:
					invalid();
				break;
			}
			command = CLEAR;
			//clear and enable pin change
			EIFR = (1 << PCIF);
			check_loop = (CHANGE_DELAY / MAIN_DELAY);
			GIMSK = (1<<PCIE);
		}
		
		//---------------------------------------------------------------------
		if ((PIND & (1<<EXT_RST))==0) {
			if (ext_rst <= 3) ext_rst++;
			if (ext_rst == 3) {
				master_reset = 1;	
				con_active();
			}
		} else {
			if (master_reset == 0) {					
				if (ext_rst >= 3) {
					PORTB = 0;
				}				
				ext_rst = 0;
			}
		}
		
		if (master_reset) {
			GIMSK = DISABLE;
			PORTB |= (master_reset);
			DDRB |= (master_reset);
			
			master_reset <<= 1;
			if (master_reset == 0) GIMSK = (1<<PCIE);
		}	

		//---------------------------------------------------------------------
		//auto check state
		if (check_loop) check_loop--;
		else
		{
			con_active();
			if (cyclic_check) 
			{
				send_port_state();
				cyclic_check = 1;
			}			
			check_loop = (AUTO_CHECK_DELAY / MAIN_DELAY);	//calculate 1 s
		}

		//---------------------------------------------------------------------
		//main delay
		 _delay_ms(MAIN_DELAY);

	}
}


//------------------------------------------
void hard_init(void)
{
	//config out port
	OUTPORT = CLEAR;
	DDRB = 0xFF;
	
	//config control signals
	SIGPORT = (1<<EXT_RST | 1<<CTS |  1<<RTS);
	DDRD = (1<<LED | 1<<TXD | 1<<CTS );
			 
	//config USART
	UCSRB = (1<<RXCIE | 1<<RXEN | 1<<TXEN);		//RX int en	
	UCSRC = (1<<UCSZ1 | 1<<UCSZ0);				//8 data, 1 stop, parity none
	UBRRL = 25;									//UBRRL = 25 - 19.2k

	//timer interrupts	
	TIMSK = (1<<TOIE1 | 1<<TOIE0);

	//pin interrupts
	PCMSK = 0xFF;
	GIMSK = (1<<PCIE);

	sei();
}

//-------------------------------------
SIGNAL(USART_RX_vect)
{
	static char index_char;
	static char temp_command;
	static char temp_parametr;
	static char temp_parametr2;
	char temp_char;
	
	temp_char = UDR;
	if (temp_char == 0x1B) index_char = 0;			//esc - new command
	if (index_char == 1) temp_command = temp_char;
	if (index_char == 2) temp_parametr = temp_char;
	if (index_char == 3) temp_parametr2 = temp_char;	
	if ((temp_char == 0x0D) && (!(command))) 		//enter - new command ready
	{
		con_active();
		if (index_char == 3) 
		{
			parametr = '0';
			parametr2 = temp_parametr;
		}
		else
		{
			parametr = temp_parametr;
			parametr2 = temp_parametr2;
		}
		command = temp_command;
	}

	if (index_char < 4) index_char++;
	if(ECHO) tx_char(temp_char);
}										

//-------------------------------------
void send_port_state(void)
{
	char port_state = PINB;
	tx_char('R');
	tx_char(get_char(port_state >> 4));
	tx_char(get_char(port_state));							
	tx_char(0x0D);	
}

//-------------------------------------
void send_pin_state(char pin)
{	
	tx_char('L');
	tx_char(get_char(pin + 1));

	if (PINB & (1<<pin))
		tx_char(get_char(SET));
	else							
		tx_char(get_char(CLEAR));
	
	tx_char(0x0D);
}

//-------------------------------------
void send_port_config(void)
{
	char port_state = DDRB;
	tx_char('G');
	tx_char(get_char(port_state >> 4));
	tx_char(get_char(port_state));							
	tx_char(0x0D);
}

//-------------------------------------
void invalid(void)
{
	tx_char('!');						
	tx_char(0x0D);
}

//-------------------------------------
void led_flasher(char pulses)
{
	pulse_index = pulses;
	TCNT1H = CLEAR;
	TCNT1L = CLEAR;
	TCCR1B = (1<<CS11 | 1<<CS11);
}

//-------------------------------------
char get_digit(char character)
{
	//digits
	if ((character >= '0') && (character <= '9'))	//0...9
		return (character - '0');
	
	if ((character >= 'A') && (character <= 'F'))	//A...F
		return ((character - 'A') + 10);	
	
	return 0xFF;
}

//-------------------------------------
char get_char(char value)
{
	value &= 0x0F;
	if (value <= 9)
		return (value + '0');
	else
		return (value + ('A' - 10));
}

//-------------------------------------
void con_active(void)
{
	led_driver(SET);
	TCNT0 = CLEAR;
	TCCR0B = (1<<CS02);// | 1<<CS00);
}

//-------------------------------------
void led_driver(char state)
{
	if (state) SIGPORT |= 1<<LED;
	else SIGPORT &= ~(1<<LED);
}

//-------------------------------------
void rx_control(char state)
{
	if (state) SIGPORT &= ~(1<<CTS);
	else SIGPORT |= (1<<CTS);
}

//-------------------------------------
void tx_char(char send)
{
	while(!(UCSRA & 1<<UDRE)){}
	UDR = send;		
	while(!(UCSRA & 1<<UDRE)){}
}

//-------------------------------------
SIGNAL(PCINT_vect)
{
	check_loop = 0;	//calculate 1/4 s
	cyclic_check = 1;
	con_active();
	check_loop = (CHANGE_DELAY / MAIN_DELAY);	//calculate 1/4 s
	_delay_ms(NOICE_DELAY);
	send_port_state();
}

//-------------------------------------
SIGNAL(TIMER1_OVF_vect)
{
	pulse_index--;
	if (pulse_index) led_driver(pulse_index & 0x01);
	else 
	{
		TCCR1B = STOP;
		led_driver(CLEAR);
	}
}

//-------------------------------------
SIGNAL(TIMER0_OVF_vect)
{
	led_driver(CLEAR);
	TCCR0B = STOP;
}




